/****************************************************************************
*
* PREFIXscatter_apnd, PREFIXscatter_back :
*		single dimension, single array
* PREFIXmultiscatter_apnd, PREFIXmultiscatter_back :
*		multi-dimensional, single array
* PREFIXmultiarr_scatter_apnd :
*		single dimension, multiple arrays
*
* History : the initial iPSC/860 version on Sep/15/93
*	    the second revision for scatter_back on Oct/07/93
* Author  : Bongki Moon (bkmoon@cs.umd.edu)
*
*****************************************************************************/

#include <stdio.h>
#include <varargs.h>
#include <assert.h>
#include "parti.h"

#define		APND		1
#define		BACK		2

/* === Single dimension, single array ====================================== */

void 
PREFIXscatter_apnd(sched, data, target)
SCHED *sched;			/* schedule */
TYPE  *data,*target;		/* data arrays */
{
	int	count,i, j;
	TYPE	**blist, *sbuf, *sbufp, *bp;
	int	*permp, *self_permp;

	blist = (TYPE**)wbuffer->blist;
	sbuf = (TYPE*)wbuffer->buffer;
	for(count=0,i=0; i<PARTI_numnodes ;i++){
		if (sched->sr->fetch_n[i]==0) continue;
		if (i == PARTI_mynode) continue;
		blist[i] = &sbuf[count];
		count += sched->sr->fetch_n[i];
	} 

	permp = sched->perm;
	for(sbufp=sbuf,i=0; i < PARTI_numnodes ;i++) {
		if (sched->sr->fetch_n[i]==0) continue;
		if (i==PARTI_mynode) {
			self_permp = permp;
			permp += sched->sr->fetch_n[PARTI_mynode];
			continue;
		}
		for(j=0; j < sched->sr->fetch_n[i] ;j++)
			*sbufp++ = data[*permp++];
	}

	for(bp=target,j=0; j < sched->sr->fetch_n[PARTI_mynode] ;j++)
		*bp++ = data[*self_permp++];

	execute_scatter_move(sched->sr, blist, target, sizeof(TYPE), APND);
}

void 
PREFIXscatter_back(sched, data, target)
SCHED *sched;			/* schedule */
TYPE  *data,*target;		/* data arrays */
{
	int	count,i, j;
	TYPE	**blist, *rbuf, *rbufp, *bp;
	int	*permp, *self_permp;

	blist = (TYPE**)wbuffer->blist;
	rbuf = (TYPE*)wbuffer->buffer;
	assert(sched->sr->total_fetch_size * sizeof(TYPE) <= wbuffer->max_size);

	for(count=0,i=0; i<PARTI_numnodes ;i++){
		if (sched->sr->fetch_n[i]==0) continue;
		if (i == PARTI_mynode) continue;
		blist[i] = &rbuf[count];
		count += sched->sr->fetch_n[i];
	} 

	execute_scatter_move(sched->sr,blist,data,sizeof(TYPE),BACK);

	for(self_permp=sched->perm,i=0;  i <= PARTI_mynode ;i++)
		self_permp += sched->sr->fetch_n[i];

	bp = data + sched->sr->fetch_n[PARTI_mynode];
	for(i=0; i < sched->sr->fetch_n[PARTI_mynode] ;i++)
		target[*--self_permp] = *--bp;

	for(permp=sched->perm,rbufp=rbuf,i=0; i < PARTI_numnodes ;i++) {
		if (sched->sr->fetch_n[i]==0) continue;
		if (i==PARTI_mynode) {
			permp += sched->sr->fetch_n[PARTI_mynode];
			continue;
		}
		for(j=0; j < sched->sr->fetch_n[i] ;j++)
			target[*permp++] = *rbufp++;
	}
}

/* === Multi-dimension, single array ====================================== */

PREFIXmultiscatter_apnd(n_sched,scheds,data,target,base_shift,dialation,repetition)
int	n_sched;
SCHED	**scheds;
TYPE	*data, *target;
int	base_shift, dialation, repetition;
{
	SCHED	*sched;
	int	count,i, j, k, ns, max_size;
	TYPE	**blist, *sbuf, *sbufp, *bp;
	int	*permp, *self_permp;

	blist = (TYPE**)wbuffer->blist;
	for(max_size=0,i=0; i < n_sched ;i++)
		if (scheds[i]->sr->total_fetch_size > max_size)
			max_size = scheds[i]->sr->total_fetch_size;
	sbuf = (TYPE*)malloc(max_size * repetition * sizeof(TYPE));

	for(ns=0; ns < n_sched ;ns++) {
		sched = scheds[ns];
		for(count=0,i=0; i<PARTI_numnodes ;i++){
			if (sched->sr->fetch_n[i]==0) continue;
			if (i == PARTI_mynode) continue;
			blist[i] = &sbuf[count];
			count += sched->sr->fetch_n[i] * repetition;
		}

		permp = sched->perm;
		for(sbufp=sbuf,i=0; i < PARTI_numnodes ;i++) {
			if (sched->sr->fetch_n[i]==0) continue;
			if (i==PARTI_mynode) {
				self_permp = permp;
				permp += sched->sr->fetch_n[PARTI_mynode];
				continue;
			}
			for(j=0; j < sched->sr->fetch_n[i] ;j++,permp++)
			    for(k=0; k < repetition ;k++)
				*sbufp++ = data[*permp * repetition + k];
		}

		for(bp=target,j=0; j < sched->sr->fetch_n[PARTI_mynode] ;j++,self_permp++)
			for(k=0; k < repetition ;k++)
				*bp++ = data[*self_permp * repetition + k];

		mdexecute_scatter_move(sched->sr,blist,target,
			sizeof(TYPE),base_shift,dialation,repetition,APND);
	}

	free(sbuf);
}

PREFIXmultiscatter_back(n_sched,scheds,data,target,base_shift,dialation,repetition)
int	n_sched;
SCHED	**scheds;
TYPE	*data, *target;
int	base_shift, dialation, repetition;
{
	SCHED	*sched;
	int	count,i, j, k, ns, max_size;
	TYPE	**blist, *rbuf, *rbufp, *bp, *bkp;
	int	*permp, *self_permp;

	blist = (TYPE**)wbuffer->blist;
	for(max_size=0,i=0; i < n_sched ;i++)
		if (scheds[i]->sr->total_fetch_size > max_size)
			max_size = scheds[i]->sr->total_fetch_size;
	rbuf = (TYPE*)malloc(max_size * repetition * sizeof(TYPE));

	for(ns=0; ns < n_sched ;ns++) {
		sched = scheds[ns];
		assert(sched->sr->total_fetch_size <= max_size);

		for(count=0,i=0; i<PARTI_numnodes ;i++){
			if (sched->sr->fetch_n[i]==0) continue;
			if (i == PARTI_mynode) continue;
			blist[i] = &rbuf[count];
			count += sched->sr->fetch_n[i] * repetition;
		}

		mdexecute_scatter_move(sched->sr,blist,data,
			sizeof(TYPE),base_shift,dialation,repetition,BACK);

		for(self_permp=sched->perm,i=0; i <= PARTI_mynode ;i++)
			self_permp += sched->sr->fetch_n[i];

		bkp = data + (sched->sr->fetch_n[PARTI_mynode]-1) * repetition;
		for(j=0; j < sched->sr->fetch_n[PARTI_mynode] ;j++) {
			for(bp=bkp,k=0; k < repetition ;k++)
				target[*(self_permp-1)*repetition + k] = *bp++;
			bkp -= repetition;
			self_permp--;
		}

		for(permp=sched->perm,rbufp=rbuf,i=0; i < PARTI_numnodes ;i++) {
			if (sched->sr->fetch_n[i]==0) continue;
			if (i==PARTI_mynode) {
				permp += sched->sr->fetch_n[PARTI_mynode];
				continue;
			}
			for(j=0; j < sched->sr->fetch_n[i] ;j++,permp++)
			    for(k=0; k < repetition ;k++)
				target[*permp * repetition + k] = *rbufp++;
		}
	}

	free(rbuf);
}

/* === Single dimension, multiple arrays ==================================== */

static void	PREFIXmaexecute_scatter_apnd();
static void	PREFIXnew_do_multiarr_scatter_apnd();

/*PREFIXnew_multiarr_scatter_apnd(sched,n_arrays,{src,dst}) */
void
PREFIXmultiarr_scatter_apnd(va_alist)
va_dcl
{
	va_list		pvar;
	SCHED		*sched;
	int		n_arrays, i;
	TYPE		**src, **dst;

	va_start(pvar);
	sched = va_arg(pvar,SCHED*);
	n_arrays = va_arg(pvar,int);
	src = (TYPE**)malloc(n_arrays * sizeof(TYPE*));
	dst = (TYPE**)malloc(n_arrays * sizeof(TYPE*));
	for(i=0; i < n_arrays ;i++) {
		src[i] = va_arg(pvar,TYPE*);
		dst[i] = va_arg(pvar,TYPE*);
	}
	va_end(pvar);

	PREFIXnew_do_multiarr_scatter_apnd(sched,src,dst,n_arrays);

	free(src);
	free(dst);
}

static void
PREFIXnew_do_multiarr_scatter_apnd(sched,src,dst,n_arrays)
SCHED	*sched;
TYPE	**src, **dst;
int	n_arrays;
{
	int	count,i, j, k;
	TYPE	**blist, *sbuf, *sbufp;
	int	*permp, *self_permp;

	blist = (TYPE**)wbuffer->blist;
	sbuf = (TYPE*)malloc(n_arrays*sizeof(TYPE)*sched->sr->total_fetch_size);
	for(count=0,i=0; i<PARTI_numnodes ;i++){
		if (sched->sr->fetch_n[i]==0) continue;
		if (i == PARTI_mynode) continue;
		blist[i] = &sbuf[count];
		count += n_arrays * sched->sr->fetch_n[i];
	} 

	permp = sched->perm;
	for(sbufp=sbuf,i=0; i < PARTI_numnodes ;i++) {
		if (sched->sr->fetch_n[i]==0) continue;
		if (i==PARTI_mynode) {
			self_permp = permp;
			permp += sched->sr->fetch_n[PARTI_mynode];
			continue;
		}
		for(j=0; j < sched->sr->fetch_n[i] ;j++,permp++)
			for(k=0; k < n_arrays ;k++) *sbufp++ = src[k][*permp];
	}

	for(j=0; j < sched->sr->fetch_n[PARTI_mynode] ;j++,self_permp++)
		for(k=0; k < n_arrays ;k++) dst[k][j] = src[k][*self_permp];

	PREFIXmaexecute_scatter_apnd(sched->sr,blist,dst,n_arrays,sizeof(TYPE));

	free(sbuf);
}

static void
PREFIXmaexecute_scatter_apnd(pattern,blist,dest,n_arrays,size)
struct com_pattern	*pattern;
TYPE	**blist, **dest;
int	n_arrays, size;
{
	static int call_count=0;
	int	i, j, k, p, count;
	int	*msg_id;
	TYPE	**wlist, *rbuf;
	TYPE	sync_buf[1];

	/* message buffer */
	msg_id = wbuffer->msg_id;

	/* allocate buffer for receiving messages */
	rbuf = (TYPE*)malloc(n_arrays*size*pattern->total_send_size);

	/* set up the buffer list */
	wlist = (TYPE**)wbuffer->wlist;
	for(count=0,p=0; p<PARTI_numnodes; p++) {
		if (p == PARTI_mynode) wlist[p] = (TYPE*)NULL;
		else {
			wlist[p] = &rbuf[count];
			count += pattern->send_n[p] * n_arrays;
		}
	}

	/* send and receive messages */
	for(i=1,count=0; i < PARTI_numnodes; i++) {
	    p = pattern->gcolor[i];
	    if (pattern->send_n[p]) {
		msg_id[count++] = PARTI_irecv(FORCED_SCATTER_EXECUTOR_MSG+
			call_count*MAXNODES+p,wlist[p],
			pattern->send_n[p]*size*n_arrays,p,PARTI_pid);
		PARTI_csend(SCATTER_SYNC+call_count*MAXNODES+PARTI_mynode,
			sync_buf,0,p,PARTI_pid);
	    }

	    if (pattern->fetch_n[p]) {
		PARTI_crecv(SCATTER_SYNC+call_count*MAXNODES+p,sync_buf,0,p,
			PARTI_pid);
		PARTI_bsend(FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+
			PARTI_mynode,blist[p],n_arrays*size
			*pattern->fetch_n[p],p,PARTI_pid);
	    }
	}

	/* wait messages and store data */
	for (i=1,count=0; i < PARTI_numnodes; i++) {
		p = pattern->gcolor[i];
		if (pattern->send_n[p]) PARTI_msgwait(msg_id[count++]);
	}

	/* store the buffered messages into the destination */
	for(k=0; k < n_arrays ;k++)
		blist[k] = dest[k] + pattern->fetch_n[PARTI_mynode];
	for(i=0; i < PARTI_numnodes; i++) {
		if (i == PARTI_mynode) continue;
		for(j=0; j < pattern->send_n[i] ;j++)
			for(k=0; k < n_arrays ;k++) *blist[k]++ = *wlist[i]++;
	}
	free(rbuf);

	call_count = (++call_count > MAXCOUNT)? 0: call_count;
}

/****** WRAPPERS BEGINS ******************************************************/

void
PREFIXscatter_apnd_(sched, data, target)
#ifdef SCHED_REUSE
int  *sched;
#else
SCHED **sched;
#endif
TYPE *data, *target;
{
#ifdef SCHED_REUSE
  PREFIXscatter_apnd((SCHED *)get_sched_ptr(*sched), data, target);
#else
  PREFIXscatter_apnd(*sched, data, target);
#endif
}

void
PREFIXscatter_back_(sched, data, target)
SCHED **sched;
TYPE *data, *target;
{
	PREFIXscatter_back(*sched, data, target);
}

void
PREFIXmultiscatter_apnd_(n_sched,scheds,data,target,base_shift,dialation,repetition)
#ifdef SCHED_REUSE
int	*n_sched, *scheds;
#else
int	*n_sched;
SCHED	**scheds;
#endif
TYPE	*data, *target;
int	*base_shift, *dialation, *repetition;
{
	int	i;
	SCHED	**sched = (SCHED**)malloc(*n_sched*sizeof(SCHED*));

#ifdef SCHED_REUSE
	for(i=0; i < *n_sched ;i++) sched[i] = (SCHED*)get_sched_ptr(scheds[i]);
#else
	for(i=0; i < *n_sched ;i++) sched[i] = scheds[i];
#endif

	PREFIXmultiscatter_apnd(*n_sched,sched,data,target,
		*base_shift,*dialation,*repetition);
	free(sched);
}

void
PREFIXmultiscatter_back_(n_sched,scheds,data,target,base_shift,dialation,repetition)
int	*n_sched;
SCHED	**scheds;
TYPE	*data, *target;
int	*base_shift, *dialation, *repetition;
{
	int	i;
	SCHED	**sched = (SCHED**)malloc(*n_sched*sizeof(SCHED*));

	for(i=0; i < *n_sched ;i++) sched[i] = scheds[i];

	PREFIXmultiscatter_back(*n_sched,sched,data,target,
		*base_shift,*dialation,*repetition);
	free(sched);
}

/*PREFIXmultiarr_scatter_apnd_(sched,n_arrays,{src,dst}) */
void
PREFIXmultiarr_scatter_apnd_(va_alist)
va_dcl
{
	va_list		pvar;
#ifdef SCHED_REUSE
	int		*sched, n_arrays, i;
#else
	int		n_arrays, i;
	SCHED		**sched;
#endif
	TYPE		**src, **dst;

	va_start(pvar);
#ifdef SCHED_REUSE
	sched = va_arg(pvar,int*);
#else
	sched = va_arg(pvar,SCHED**);
#endif
	n_arrays = *va_arg(pvar,int*);
	src = (TYPE**)malloc(n_arrays * sizeof(TYPE*));
	dst = (TYPE**)malloc(n_arrays * sizeof(TYPE*));
	for(i=0; i < n_arrays ;i++) {
		src[i] = va_arg(pvar,TYPE*);
		dst[i] = va_arg(pvar,TYPE*);
	}
	va_end(pvar);

#ifdef SCHED_REUSE
	PREFIXnew_do_multiarr_scatter_apnd((SCHED *)get_sched_ptr(*sched),
		src, dst, n_arrays);
#else
	PREFIXnew_do_multiarr_scatter_apnd(*sched, src, dst, n_arrays);
#endif

	free(src);
	free(dst);
}

/****** WRAPPERS ENDS ********************************************************/
